/*********************************************************
 *
 * IMS_MIDI.c
 *
 * Authors: Ben McQuiston
 *
 * C function to convert our house IMS files into a usable MIDI file
 * 
 * Last Revision: 
 *       - October 25, 2006
 *			Major structure created for Mthd print and 
 *          first MTrk with tempo specifiers 
 *
 *       - October 29, 2006
 *			Note processing code added, to decide how many bytes
 *			a given track needs to be, and write the actual delta
 *			times and note values to the MIDI file.
 *  
 *****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int
ImsToMIDI(char *filename)
{
	printf("Generating MIDI file...\n");
	
	int tempo;
	int divisor = 60000000;
	int mspqn;
	char zero = 0x00;

	long fileMrk;
	int numTrkByts = 0;

	FILE *fp1;
	FILE *fp2;

	char ch;
	char keySig;
	char timeSigTop;
	char timeSigBot;
	char instr;
	char noteVal;

	fp1 = fopen(filename, "rb");
	fp2 = fopen("output.mid","w");

	if(fp1 == NULL){
	  printf("ERROR\n");
	  return -1;
	}

	
/********************************************************
*
*  Mthd header to file...this will always be the same
*
*********************************************************/
		
	fprintf(fp2,"%c%c%c%c",0x4D,0x54,0x68,0x64); /* Mthd */
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",0x06);/* Always 6 bytes in Mthd header */
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",0x01);/* MIDI Type 1 */
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",0x02);/* Number of Tracks... could change?*/
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",0xF0); /*240 PPQN always..so dont forget*/

	 /*fprintf(fp2,"%c",zero); delta time 00 */

/********************************************************
*  
*   Mtrk Header...
*
*********************************************************/

	fprintf(fp2,"%c%c%c%c",0x4D,0x54,0x72,0x6B); /* Mtrk */
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c",0x19); /* 25 bytes in first track...stays the same */

	fprintf(fp2,"%c",zero); /* delta time 00 */
	
/******************************************************
*
*  Start at beginning of IMS file to start getting 
*  musical information...
*
********************************************************/

	fseek(fp1,0,SEEK_SET);
	
	/* tempo cammand FF 51 03 xx xx xx = micros / quarter note */

	fscanf(fp1,"%c",&ch);
	tempo = (int) ch;
	tempo = tempo & ~(0xFFFFFFFF << 8);/* mask off all but the lower byte); */

	mspqn = divisor / tempo;
	fprintf(fp2,"%c%c%c", 0xFF,0x51,0x03); /* tempo command */
	fprintf(fp2,"%c%c%c",*((char *) &mspqn +2),*((char *) &mspqn+1), *((char *) &mspqn));

	fprintf(fp2,"%c",zero);  /*delta time 00 */

	/* key sig command FF 59 02 xx xx = flats or sharps and major or minor */

	fscanf(fp1,"%c",&ch);
	keySig = ch;

	/* logic........for now we are just going to force it 
	* to C major no flats we are not sure why we need this...*/
	
	fprintf(fp2,"%c%c%c",0xFF,0x59,0x02);
	fprintf(fp2,"%c",zero); /* no flats (C) */
	fprintf(fp2,"%c",zero); /* major */ 
	fprintf(fp2,"%c",zero); /* delta time 00 */

	/* Time Signature command FF 58 04 xx xx xx xx 
	* numerator and denominator of time signature
	* number of MIDI clocks per beat (always 24)
	* number of 32nd notes per quater note (always 8)
	*/
	
	fscanf(fp1,"%c",&ch);
	timeSigTop = ch;
	fprintf(fp2,"%c%c%c",0xFF,0x58,0x04);

	fprintf(fp2,"%c",ch);

	fscanf(fp1,"%c",&ch);
	timeSigBot = ch;

	/* denominator must be a negative power of four,
	*  so we check if its a 4 or an 8 and print the corresponding
	* power to the file...
	*/

	if ((int) ch == 4){
		fprintf(fp2,"%c",0x02); 
	}
	else if ((int) ch == 8){
		fprintf(fp2,"%c",0x03);
	}

	fprintf(fp2,"%c%c",0x18,0x08); /* 24 MIDI clocks per beat,
								   * 8 32nds per quarter*/								   

	fprintf(fp2,"%c",zero);  /* delta time 00 */

	fprintf(fp2,"%c%c%c",0xFF,0x2F,0); /* End of Track Marker */


/*************************************************************
*
*  NOTE PROCESSING
*
*  Actually deal with the note values and lengths
*  Figure out how long it is going to be so that the Mtrk
*  header can tell how many bytes to expect, then just 
*  process all of the notes two bytes at a time...
*
************************************************************/
	
	fprintf(fp2,"%c%c%c%c",0x4D,0x54,0x72,0x6B); /* Mtrk */
	
	/* So to make this work we have to do an initial search through
	* the file to figure out what is in there...basically to figure out
	* out how many bytes are goiung to be in the file....*/


	fscanf(fp1,"%c",&ch);
	instr = ch;

	fileMrk = 5; 

	while (fscanf(fp1,"%c",&ch) == 1){
		
		fscanf(fp1,"%c",&ch);
		if (ch == 0x03 || ch == 0x04 || ch == 0x05){
			numTrkByts += 8;
		}
		else{
			numTrkByts += 9;
		}
	}
	
	printf("count: %d\n", numTrkByts);

	fseek(fp1,fileMrk,SEEK_SET);

	numTrkByts += 8; /* + 4 to account for the four bytes to tell
					 * what instrument is to be played and +4 for
					 the end of track bytes*/

	fprintf(fp2,"%c",*((char *) &numTrkByts + 3)); 
	fprintf(fp2,"%c",*((char *) &numTrkByts + 2));
	fprintf(fp2,"%c",*((char *) &numTrkByts + 1));
	fprintf(fp2,"%c",*((char *) &numTrkByts));	/* actual number of bytes in
												* the coming track */
	
	/*After we figure out how many bytes there will be in the track,
	* we have to actually process the notes and things...*/

	fprintf(fp2,"%c",zero);
	fprintf(fp2,"%c%c",0xC0,instr); /*tells what instrument it is ..*/
	fprintf(fp2,"%c",zero);


	while (fscanf(fp1,"%c",&ch) == 1){
		noteVal = ch;
		fprintf(fp2,"%c%c%c",0x90,noteVal,0x40);
		fscanf(fp1,"%c",&ch);
		if (ch == 0){ /* whole note = 960 pulses */
			fprintf(fp2,"%c",0x87);
			fprintf(fp2,"%c",0x40);
		}
		else if (ch == 0x01){ /* half note = 480 pulses */
			fprintf(fp2,"%c",0x83);
			fprintf(fp2,"%c",0x60);
		}
		else if (ch == 0x02){ /* quarter note = 240 pulses */
			fprintf(fp2,"%c",0x81);
			fprintf(fp2,"%c",0x70);
		}
		else if (ch == 0x03){ /* eigth note = 120 pulses */
			fprintf(fp2,"%c",0x78);
		}
		else if (ch == 0x04){ /* 16th note = 60 pulses */
			fprintf(fp2,"%c",0x3C);
		}
		else if (ch == 0x05){/* 32nd note = 30 pulses */
			fprintf(fp2,"%c",0x1E);
		}
		else if (ch == 0x06){/* dotted eigth note = 180 pulses */
			fprintf(fp2,"%c",0x81);
			fprintf(fp2,"%c",0x34);
		}
		else if (ch == 0x07){ /* dotted quarter note = 360 pulses */
			fprintf(fp2,"%c",0x82);
			fprintf(fp2,"%c",0x68);
		}
		else if (ch == 0x08){ /* dotted half note = 720 pulses */
			fprintf(fp2,"%c",0x83);
			fprintf(fp2,"%c",0x50);
		}

		fprintf(fp2,"%c%c%c",0x80,noteVal,0x40);
		fprintf(fp2,"%c",zero);
	}

	fseek(fp2,-1,SEEK_CUR);

	fprintf(fp2,"%c",0xFF);
	fprintf(fp2,"%c",0x7F);
	fprintf(fp2,"%c%c%c",0xFF,0x2F,0); /* End of Track Marker */

	return 0;
	
}
